(* convolutions *)
let node convo (c0, c1, c2) = (c0 + c1 + c2) / 3

let node convolution (p0, p1, p2) = p where
  rec p = (r,g,b)
  and r = convo (p0r, p1r, p2r) 
  and g = convo (p0g, p1g, p2g)
  and b = convo (p0b, p1b, p2b) 
  and p0r, p0g, p0b = p0
  and p1r, p1g, p1b = p1
  and p2r, p2g, p2b = p2

(* horizontal filter *)
let clock hf = (10100100) 

let node horizontal_filter p = o where
  rec p0 = p fby p1
  and p1 = p fby p2
  and p2 = p 
  and o = (convolution (p0, p1, p2)) when hf

(* vertical sliding_window *)
let clock first_sd_line = 1^720 (0)
let clock first_line_of_img = (1^720 0^(720*1079))
let clock last_line_of_img =  (0^(720*1079) 1^720)

let node my_fby_sd_line (p1,p2) = 
  merge first_sd_line (p1 when first_sd_line) (buffer(p2))

let node reorder p = ((p0,p1,p2)::'a) where
  rec p0 = 
    merge first_line_of_img 
      (p1 when first_line_of_img)
      ((my_fby_sd_line (p1, p1)) whenot first_line_of_img)
  and p1 = buffer(p)
  and p2 = 
    merge last_line_of_img 
      (p1 when last_line_of_img)
      ((p whenot first_sd_line) whenot last_line_of_img) 

(* vertical filter *)
let clock vf = (1^720 0^720 1^720 0^720 0^720 1^720 0^720 0^720 1^720)

let node vertical_filter p = o where
  rec (p0,p1,p2) = reorder p
  and o = (convolution (p0, p1, p2)) when vf

(* downscaler *)
let node downscaler p = vertical_filter (horizontal_filter p)

(* picture in picture *)
let clock incrust_end = 
  (0^(1920 * (1080 - 480)) {0^1200 1^720}^480)

let node picture_in_picture_end (p1, p2) = o where
  rec small = buffer(downscaler p1)
  and big = (p2 whenot incrust_end)
  and o = merge incrust_end small big

(* picture in picture *)
let clock incrust_begin = 
  ({1^720 0^1200}^480 0^(1920 * (1080 - 480)))

let node picture_in_picture_begin (p1, p2) = o where
  rec small = buffer(downscaler p1)
  and big = p2 whenot incrust_begin
  and o = merge incrust_begin small big
